@@ -2,15 +2,21 @@ use std::convert::TryInto;
 use std::mem::ManuallyDrop;
 use std::ops::DerefMut;
 
-use crate::type_conversions::insert_at;
+use crate::shared_types::{
+    DeepSubscription, DefaultPyErr, PreliminaryObservationException, ShallowSubscription, SubId,
+};
+use crate::type_conversions::events_into_py;
 use crate::y_transaction::YTransaction;
 
 use super::shared_types::SharedType;
 use crate::type_conversions::ToPython;
-use pyo3::exceptions::{PyIndexError, PyTypeError};
+use pyo3::exceptions::PyIndexError;
+
+use crate::type_conversions::{py_into_any, PyObjectWrapper};
 use pyo3::prelude::*;
 use pyo3::types::{PyList, PySlice, PySliceIndices};
 use yrs::types::array::{ArrayEvent, ArrayIter};
+use yrs::types::DeepObservable;
 use yrs::{Array, SubscriptionId, Transaction};
 
 /// A collection used to store data in an indexed sequence structure. This type is internally
@@ -50,26 +56,7 @@ impl YArray {
     /// document store and cannot be nested again: attempt to do so will result in an exception.
     #[new]
     pub fn new(init: Option<PyObject>) -> PyResult<Self> {
-        let elements = if let Some(iterable) = init {
-            Python::with_gil(|py| {
-                iterable.as_ref(py).iter().map(|iterable| {
-                    iterable
-                        .map(|element| match element {
-                            Ok(value) => {
-                                let obj: PyObject = value.into();
-                                obj
-                            }
-                            Err(py_err) => {
-                                py_err.restore(py);
-                                py.None()
-                            }
-                        })
-                        .collect()
-                })
-            })
-        } else {
-            Ok(vec![])
-        };
+        let elements = init.map(Self::py_iter).unwrap_or(Ok(Vec::default()));
         elements.map(|el_array| YArray(SharedType::prelim(el_array)))
     }
 
@@ -112,32 +99,74 @@ impl YArray {
             }
         })
     }
+    /// Adds a single item to the provided index in the array.
+    pub fn insert(&mut self, txn: &mut YTransaction, index: u32, item: PyObject) -> PyResult<()> {
+        match &mut self.0 {
+            SharedType::Integrated(array) if array.len() >= index => {
+                py_into_any(item).map(|any| array.insert(txn, index, any))
+            }
+            SharedType::Prelim(vec) if vec.len() >= index as usize => {
+                Ok(vec.insert(index as usize, item))
+            }
+            _ => Err(PyIndexError::default_message()),
+        }
+    }
 
     /// Inserts a given range of `items` into this `YArray` instance, starting at given `index`.
-    pub fn insert(&mut self, txn: &mut YTransaction, index: u32, items: Vec<PyObject>) {
-        let mut j = index;
+    pub fn insert_range(
+        &mut self,
+        txn: &mut YTransaction,
+        index: u32,
+        items: PyObject,
+    ) -> PyResult<()> {
+        let items = Self::py_iter(items)?;
         match &mut self.0 {
-            SharedType::Integrated(array) => {
-                insert_at(array, txn, index, items);
+            SharedType::Integrated(array) if array.len() >= index => {
+                Self::insert_multiple_at(array, txn, index, items);
+                Ok(())
             }
-            SharedType::Prelim(vec) => {
+            SharedType::Prelim(vec) if vec.len() >= index as usize => {
+                let mut j = index;
                 for el in items {
                     vec.insert(j as usize, el);
                     j += 1;
                 }
+                Ok(())
             }
+            _ => Err(PyIndexError::default_message()),
         }
     }
 
     /// Appends a range of `items` at the end of this `YArray` instance.
-    pub fn push(&mut self, txn: &mut YTransaction, items: Vec<PyObject>) {
-        let index = self.__len__() as u32; // length guaranteed to be non negative
-        self.insert(txn, index, items);
+    pub fn extend(&mut self, txn: &mut YTransaction, items: PyObject) -> PyResult<()> {
+        let index = self.__len__() as u32;
+        self.insert_range(txn, index, items)
+    }
+    /// Adds a single item to the end of the array
+    pub fn append(&mut self, txn: &mut YTransaction, item: PyObject) {
+        match &mut self.0 {
+            SharedType::Integrated(array) => {
+                let wrapper = PyObjectWrapper(item);
+                array.push_back(txn, wrapper)
+            }
+            SharedType::Prelim(vec) => vec.push(item),
+        }
+    }
+    /// Removes the element that the given index from the list.
+    pub fn delete(&mut self, txn: &mut YTransaction, index: u32) -> PyResult<()> {
+        match &mut self.0 {
+            SharedType::Integrated(v) if index < v.len() => Ok(v.remove(txn, index)),
+            SharedType::Prelim(v) if index < v.len() as u32 => {
+                v.remove(index as usize);
+                Ok(())
+            }
+            _ => Err(PyIndexError::default_message()),
+        }
     }
 
     /// Deletes a range of items of given `length` from current `YArray` instance,
     /// starting from given `index`.
-    pub fn delete(&mut self, txn: &mut YTransaction, index: u32, length: u32) {
+    pub fn delete_range(&mut self, txn: &mut YTransaction, index: u32, length: u32) {
         match &mut self.0 {
             SharedType::Integrated(v) => v.remove_range(txn, index, length),
             SharedType::Prelim(v) => {
@@ -186,35 +215,52 @@ impl YArray {
     /// Subscribes to all operations happening over this instance of `YArray`. All changes are
     /// batched and eventually triggered during transaction commit phase.
     /// Returns a `SubscriptionId` which can be used to cancel the callback with `unobserve`.
-    pub fn observe(&mut self, f: PyObject) -> PyResult<SubscriptionId> {
+    pub fn observe(&mut self, f: PyObject) -> PyResult<ShallowSubscription> {
         match &mut self.0 {
             SharedType::Integrated(array) => {
-                let subscription = array.observe(move |txn, e| {
-                    Python::with_gil(|py| {
-                        let event = YArrayEvent::new(e, txn);
-                        if let Err(err) = f.call1(py, (event,)) {
-                            err.restore(py)
-                        }
+                let sub: SubscriptionId = array
+                    .observe(move |txn, e| {
+                        Python::with_gil(|py| {
+                            let event = YArrayEvent::new(e, txn);
+                            if let Err(err) = f.call1(py, (event,)) {
+                                err.restore(py)
+                            }
+                        })
                     })
-                });
-                Ok(subscription.into())
+                    .into();
+                Ok(ShallowSubscription(sub))
             }
-            SharedType::Prelim(_) => Err(PyTypeError::new_err(
-                "Cannot observe a preliminary type. Must be added to a YDoc first",
-            )),
+            SharedType::Prelim(_) => Err(PreliminaryObservationException::default_message()),
+        }
+    }
+    /// Observes YArray events and events of all child elements.
+    pub fn observe_deep(&mut self, f: PyObject) -> PyResult<DeepSubscription> {
+        match &mut self.0 {
+            SharedType::Integrated(array) => {
+                let sub: SubscriptionId = array
+                    .observe_deep(move |txn, events| {
+                        Python::with_gil(|py| {
+                            let events = events_into_py(txn, events);
+                            if let Err(err) = f.call1(py, (events,)) {
+                                err.restore(py)
+                            }
+                        })
+                    })
+                    .into();
+                Ok(DeepSubscription(sub))
+            }
+            SharedType::Prelim(_) => Err(PreliminaryObservationException::default_message()),
         }
     }
 
     /// Cancels the callback of an observer using the Subscription ID returned from the `observe` method.
-    pub fn unobserve(&mut self, subscription_id: SubscriptionId) -> PyResult<()> {
+    pub fn unobserve(&mut self, subscription_id: SubId) -> PyResult<()> {
         match &mut self.0 {
-            SharedType::Integrated(v) => {
-                v.unobserve(subscription_id);
-                Ok(())
-            }
-            SharedType::Prelim(_) => Err(PyTypeError::new_err(
-                "Cannot call unobserve on a preliminary type. Must be added to a YDoc first",
-            )),
+            SharedType::Integrated(arr) => Ok(match subscription_id {
+                SubId::Shallow(ShallowSubscription(id)) => arr.unobserve(id),
+                SubId::Deep(DeepSubscription(id)) => arr.unobserve_deep(id),
+            }),
+            SharedType::Prelim(_) => Err(PreliminaryObservationException::default_message()),
         }
     }
 }
@@ -227,18 +273,14 @@ impl YArray {
                 if let Some(value) = v.get(index as u32) {
                     Ok(Python::with_gil(|py| value.into_py(py)))
                 } else {
-                    Err(PyIndexError::new_err(
-                        "Index outside the bounds of an YArray",
-                    ))
+                    Err(PyIndexError::default_message())
                 }
             }
             SharedType::Prelim(v) => {
                 if let Some(value) = v.get(index as usize) {
                     Ok(value.clone())
                 } else {
-                    Err(PyIndexError::new_err(
-                        "Index outside the bounds of an YArray",
-                    ))
+                    Err(PyIndexError::default_message())
                 }
             }
         }
@@ -301,6 +343,48 @@ impl YArray {
             index as u32
         }
     }
+
+    pub fn insert_multiple_at(dst: &Array, txn: &mut Transaction, index: u32, src: Vec<PyObject>) {
+        let mut j = index;
+        let mut i = 0;
+        while i < src.len() {
+            let mut anys = Vec::default();
+            while i < src.len() {
+                if let Ok(any) = py_into_any(src[i].clone()) {
+                    anys.push(any);
+                    i += 1;
+                } else {
+                    break;
+                }
+            }
+
+            if !anys.is_empty() {
+                let len = anys.len() as u32;
+                dst.insert_range(txn, j, anys);
+                j += len;
+            } else {
+                let wrapper = PyObjectWrapper(src[i].clone());
+                dst.insert(txn, j, wrapper);
+                i += 1;
+                j += 1;
+            }
+        }
+    }
+
+    fn py_iter(iterable: PyObject) -> PyResult<Vec<PyObject>> {
+        Python::with_gil(|py| {
+            iterable.as_ref(py).iter().and_then(|iterable| {
+                iterable
+                    .map(|element| {
+                        element.map(|el| {
+                            let el: PyObject = el.into();
+                            el
+                        })
+                    })
+                    .collect()
+            })
+        })
+    }
 }
 #[derive(FromPyObject)]
 pub enum Index<'a> {
@@ -356,7 +440,7 @@ pub struct YArrayEvent {
 }
 
 impl YArrayEvent {
-    fn new(event: &ArrayEvent, txn: &Transaction) -> Self {
+    pub fn new(event: &ArrayEvent, txn: &Transaction) -> Self {
         let inner = event as *const ArrayEvent;
         let txn = txn as *const Transaction;
         YArrayEvent {
@@ -421,3 +505,9 @@ impl YArrayEvent {
         }
     }
 }
+
+impl DefaultPyErr for PyIndexError {
+    fn default_message() -> PyErr {
+        PyIndexError::new_err("Index out of bounds.")
+    }
+}
