@@ -1,13 +1,19 @@
+use crate::shared_types::{
+    DeepSubscription, DefaultPyErr, IntegratedOperationException, PreliminaryObservationException,
+    ShallowSubscription, SharedType, SubId,
+};
+use crate::type_conversions::py_into_any;
+use crate::type_conversions::{events_into_py, ToPython};
+use crate::y_transaction::YTransaction;
 use lib0::any::Any;
-use pyo3::exceptions::PyTypeError;
 use pyo3::prelude::*;
 use pyo3::types::PyList;
+use std::collections::HashMap;
+use std::rc::Rc;
 use yrs::types::text::TextEvent;
-use yrs::{SubscriptionId, Text, Transaction};
-
-use crate::shared_types::SharedType;
-use crate::type_conversions::ToPython;
-use crate::y_transaction::YTransaction;
+use yrs::types::Attrs;
+use yrs::types::DeepObservable;
+use yrs::{Text, Transaction};
 
 /// A shared data type used for collaborative text editing. It enables multiple users to add and
 /// remove chunks of text in efficient manner. This type is internally represented as a mutable
@@ -86,24 +92,99 @@ impl YText {
     }
 
     /// Inserts a given `chunk` of text into this `YText` instance, starting at a given `index`.
-    pub fn insert(&mut self, txn: &mut YTransaction, index: u32, chunk: &str) {
+    pub fn insert(
+        &mut self,
+        txn: &mut YTransaction,
+        index: u32,
+        chunk: &str,
+        attributes: Option<HashMap<String, PyObject>>,
+    ) -> PyResult<()> {
+        let attributes: Option<PyResult<Attrs>> = attributes.map(Self::parse_attrs);
+
+        if let Some(Ok(attributes)) = attributes {
+            match &mut self.0 {
+                SharedType::Integrated(text) => {
+                    text.insert_with_attributes(txn, index, chunk, attributes);
+                    Ok(())
+                }
+                SharedType::Prelim(_) => Err(IntegratedOperationException::default_message()),
+            }
+        } else if let Some(Err(error)) = attributes {
+            Err(error)
+        } else {
+            match &mut self.0 {
+                SharedType::Integrated(text) => text.insert(txn, index, chunk),
+                SharedType::Prelim(prelim_string) => {
+                    prelim_string.insert_str(index as usize, chunk)
+                }
+            }
+            Ok(())
+        }
+    }
+
+    /// Inserts a given `embed` object into this `YText` instance, starting at a given `index`.
+    ///
+    /// Optional object with defined `attributes` will be used to wrap provided `embed`
+    /// with a formatting blocks.`attributes` are only supported for a `YText` instance which
+    /// already has been integrated into document store.
+    pub fn insert_embed(
+        &mut self,
+        txn: &mut YTransaction,
+        index: u32,
+        embed: PyObject,
+        attributes: Option<HashMap<String, PyObject>>,
+    ) -> PyResult<()> {
         match &mut self.0 {
-            SharedType::Integrated(v) => v.insert(txn, index, chunk),
-            SharedType::Prelim(v) => v.insert_str(index as usize, chunk),
+            SharedType::Integrated(text) => {
+                let content = py_into_any(embed)?;
+                if let Some(Ok(attrs)) = attributes.map(Self::parse_attrs) {
+                    text.insert_embed_with_attributes(txn, index, content, attrs)
+                } else {
+                    text.insert_embed(txn, index, content)
+                }
+                Ok(())
+            }
+            SharedType::Prelim(_) => Err(IntegratedOperationException::default_message()),
+        }
+    }
+
+    /// Wraps an existing piece of text within a range described by `index`-`length` parameters with
+    /// formatting blocks containing provided `attributes` metadata. This method only works for
+    /// `YText` instances that already have been integrated into document store.
+    pub fn format(
+        &mut self,
+        txn: &mut YTransaction,
+        index: u32,
+        length: u32,
+        attributes: HashMap<String, PyObject>,
+    ) -> PyResult<()> {
+        match Self::parse_attrs(attributes) {
+            Ok(attrs) => match &mut self.0 {
+                SharedType::Integrated(text) => {
+                    text.format(txn, index, length, attrs);
+                    Ok(())
+                }
+                SharedType::Prelim(_) => Err(IntegratedOperationException::default_message()),
+            },
+            Err(err) => Err(err),
         }
     }
 
     /// Appends a given `chunk` of text at the end of current `YText` instance.
-    pub fn push(&mut self, txn: &mut YTransaction, chunk: &str) {
+    pub fn extend(&mut self, txn: &mut YTransaction, chunk: &str) {
         match &mut self.0 {
             SharedType::Integrated(v) => v.push(txn, chunk),
             SharedType::Prelim(v) => v.push_str(chunk),
         }
     }
+    /// Deletes character at the specified index.
+    pub fn delete(&mut self, txn: &mut YTransaction, index: u32) {
+        self.delete_range(txn, index, 1);
+    }
 
     /// Deletes a specified range of of characters, starting at a given `index`.
     /// Both `index` and `length` are counted in terms of a number of UTF-8 character bytes.
-    pub fn delete(&mut self, txn: &mut YTransaction, index: u32, length: u32) {
+    pub fn delete_range(&mut self, txn: &mut YTransaction, index: u32, length: u32) {
         match &mut self.0 {
             SharedType::Integrated(v) => v.remove_range(txn, index, length),
             SharedType::Prelim(v) => {
@@ -112,37 +193,70 @@ impl YText {
         }
     }
 
-    pub fn observe(&mut self, f: PyObject) -> PyResult<SubscriptionId> {
+    /// Observes updates from the `YText` instance.
+    pub fn observe(&mut self, f: PyObject) -> PyResult<ShallowSubscription> {
         match &mut self.0 {
-            SharedType::Integrated(v) => Ok(v
-                .observe(move |txn, e| {
-                    Python::with_gil(|py| {
-                        let e = YTextEvent::new(e, txn);
-                        if let Err(err) = f.call1(py, (e,)) {
-                            err.restore(py)
-                        }
-                    });
-                })
-                .into()),
-            SharedType::Prelim(_) => Err(PyTypeError::new_err(
-                "Cannot observe a preliminary type. Must be added to a YDoc first",
-            )),
+            SharedType::Integrated(text) => {
+                let sub_id = text
+                    .observe(move |txn, e| {
+                        Python::with_gil(|py| {
+                            let e = YTextEvent::new(e, txn);
+                            if let Err(err) = f.call1(py, (e,)) {
+                                err.restore(py)
+                            }
+                        });
+                    })
+                    .into();
+                Ok(ShallowSubscription(sub_id))
+            }
+            SharedType::Prelim(_) => Err(PreliminaryObservationException::default_message()),
+        }
+    }
+
+    /// Observes updates from the `YText` instance and all of its nested children.
+    pub fn observe_deep(&mut self, f: PyObject) -> PyResult<DeepSubscription> {
+        match &mut self.0 {
+            SharedType::Integrated(text) => {
+                let sub = text
+                    .observe_deep(move |txn, events| {
+                        Python::with_gil(|py| {
+                            let events = events_into_py(txn, events);
+                            if let Err(err) = f.call1(py, (events,)) {
+                                err.restore(py)
+                            }
+                        })
+                    })
+                    .into();
+                Ok(DeepSubscription(sub))
+            }
+            SharedType::Prelim(_) => Err(PreliminaryObservationException::default_message()),
         }
     }
     /// Cancels the observer callback associated with the `subscripton_id`.
-    pub fn unobserve(&mut self, subscription_id: SubscriptionId) -> PyResult<()> {
+    pub fn unobserve(&mut self, subscription_id: SubId) -> PyResult<()> {
         match &mut self.0 {
-            SharedType::Integrated(text) => {
-                text.unobserve(subscription_id);
-                Ok(())
-            }
-            SharedType::Prelim(_) => Err(PyTypeError::new_err(
-                "Cannot unobserve a preliminary type. Must be added to a YDoc first",
-            )),
+            SharedType::Integrated(text) => Ok(match subscription_id {
+                SubId::Shallow(ShallowSubscription(id)) => text.unobserve(id),
+                SubId::Deep(DeepSubscription(id)) => text.unobserve_deep(id),
+            }),
+            SharedType::Prelim(_) => Err(PreliminaryObservationException::default_message()),
         }
     }
 }
 
+impl YText {
+    fn parse_attrs(attrs: HashMap<String, PyObject>) -> PyResult<Attrs> {
+        attrs
+            .into_iter()
+            .map(|(k, v)| {
+                let key = Rc::from(k);
+                let value = py_into_any(v)?;
+                Ok((key, value))
+            })
+            .collect()
+    }
+}
+
 /// Event generated by `YYText.observe` method. Emitted during transaction commit phase.
 #[pyclass(unsendable)]
 pub struct YTextEvent {
@@ -153,7 +267,7 @@ pub struct YTextEvent {
 }
 
 impl YTextEvent {
-    fn new(event: &TextEvent, txn: &Transaction) -> Self {
+    pub fn new(event: &TextEvent, txn: &Transaction) -> Self {
         let inner = event as *const TextEvent;
         let txn = txn as *const Transaction;
         YTextEvent {
@@ -218,4 +332,15 @@ impl YTextEvent {
             delta
         }
     }
+
+    fn __str__(&self) -> String {
+        format!(
+            "YTextEvent(target={:?}, delta={:?})",
+            self.target, self.delta
+        )
+    }
+
+    fn __repr__(&self) -> String {
+        self.__str__()
+    }
 }
